@@ -1,24 +1,28 @@
 % Licensed under the Apache License, Version 2.0 (the "License"); you may not
-% use this file except in compliance with the License.  You may obtain a copy of
+% use this file except in compliance with the License. You may obtain a copy of
 % the License at
 %
 %   http://www.apache.org/licenses/LICENSE-2.0
 %
 % Unless required by applicable law or agreed to in writing, software
 % distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
-% WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
+% WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 % License for the specific language governing permissions and limitations under
 % the License.
 
 -module(couch_util).
 
--export([start_driver/1]).
--export([should_flush/0, should_flush/1, to_existing_atom/1, to_binary/1]).
+-export([start_driver/1,terminate_linked/1]).
+-export([should_flush/0, should_flush/1, to_existing_atom/1]).
 -export([new_uuid/0, rand32/0, implode/2, collate/2, collate/3]).
 -export([abs_pathname/1,abs_pathname/2, trim/1, ascii_lower/1]).
--export([encodeBase64/1, decodeBase64/1, to_hex/1,parse_term/1,dict_find/3]).
+-export([encodeBase64/1, decodeBase64/1, encodeBase64Url/1, decodeBase64Url/1,
+    to_hex/1,parse_term/1, dict_find/3]).
+-export([file_read_size/1, get_nested_json_value/2, json_user_ctx/1]).
+-export([to_binary/1, to_list/1, url_encode/1]).
 
 -include("couch_db.hrl").
+-include_lib("kernel/include/file.hrl").
 
 % arbitrarily chosen amount of memory to use before flushing to disk
 -define(FLUSH_MAX_MEM, 10000000).
@@ -43,9 +47,17 @@ to_existing_atom(V) when is_atom(V)->
     V.
 
 
+terminate_linked(normal) ->
+    terminate_linked(shutdown);
+terminate_linked(Reason) ->
+    {links, Links} = process_info(self(), links),
+    [catch exit(Pid, Reason) || Pid <- Links],
+    ok.
+
+
 new_uuid() ->
     list_to_binary(to_hex(crypto:rand_bytes(16))).
-    
+
 to_hex([]) ->
     [];
 to_hex(Bin) when is_binary(Bin) ->
@@ -57,24 +69,27 @@ to_digit(N) when N < 10 -> $0 + N;
 to_digit(N)             -> $a + N-10.
 
 
-to_binary(V) when is_binary(V) ->
-    V;
-to_binary(V) when is_list(V) -> 
-    try list_to_binary(V)
-    catch
-        _ -> list_to_binary(io_lib:format("~p", [V]))
-    end;
-to_binary(V) when is_atom(V) ->
-    list_to_binary(atom_to_list(V));
-to_binary(V) ->
-    list_to_binary(io_lib:format("~p", [V])).
-
-
 parse_term(Bin) when is_binary(Bin)->
     parse_term(binary_to_list(Bin));
 parse_term(List) ->
     {ok, Tokens, _} = erl_scan:string(List ++ "."),
     erl_parse:parse_term(Tokens).
+
+
+get_nested_json_value({Props}, [Key|Keys]) ->
+    case proplists:get_value(Key, Props, nil) of
+    nil -> throw({not_found, <<"missing json key: ", Key/binary>>});
+    Value -> get_nested_json_value(Value, Keys)
+    end;
+get_nested_json_value(Value, []) ->
+    Value;
+get_nested_json_value(_NotJSONObj, _) ->
+    throw({not_found, json_mismatch}).
+
+json_user_ctx(#db{name=DbName, user_ctx=Ctx}) ->
+    {[{<<"db">>, DbName},
+            {<<"name">>,Ctx#user_ctx.name},
+            {<<"roles">>,Ctx#user_ctx.roles}]}.
     
 
 % returns a random integer
@@ -195,15 +210,15 @@ collate(A, B, Options) when is_binary(A), is_binary(B) ->
 
 should_flush() ->
     should_flush(?FLUSH_MAX_MEM).
-    
+
 should_flush(MemThreshHold) ->
     {memory, ProcMem} = process_info(self(), memory),
-    BinMem = lists:foldl(fun({_Id, Size, _NRefs}, Acc) -> Size+Acc end, 
+    BinMem = lists:foldl(fun({_Id, Size, _NRefs}, Acc) -> Size+Acc end,
         0, element(2,process_info(self(), binary))),
     if ProcMem+BinMem > 2*MemThreshHold ->
         garbage_collect(),
         {memory, ProcMem2} = process_info(self(), memory),
-        BinMem2 = lists:foldl(fun({_Id, Size, _NRefs}, Acc) -> Size+Acc end, 
+        BinMem2 = lists:foldl(fun({_Id, Size, _NRefs}, Acc) -> Size+Acc end,
             0, element(2,process_info(self(), binary))),
         if ProcMem2+BinMem2 > MemThreshHold ->
             true;
@@ -228,11 +243,11 @@ should_flush(MemThreshHold) ->
 %% Take 3 bytes a time (3 x 8 = 24 bits), and make 4 characters out of
 %% them (4 x 6 = 24 bits).
 %%
-encodeBase64(Bs) when list(Bs) ->
-    encodeBase64(list_to_binary(Bs), <<>>);
+encodeBase64(Bs) when is_list(Bs) ->
+    encodeBase64(iolist_to_binary(Bs), <<>>);
 encodeBase64(Bs) ->
     encodeBase64(Bs, <<>>).
-    
+
 encodeBase64(<<B:3/binary, Bs/binary>>, Acc) ->
     <<C1:6, C2:6, C3:6, C4:6>> = B,
     encodeBase64(Bs, <<Acc/binary, (enc(C1)), (enc(C2)), (enc(C3)), (enc(C4))>>);
@@ -245,6 +260,23 @@ encodeBase64(<<B:1/binary>>, Acc) ->
 encodeBase64(<<>>, Acc) ->
     Acc.
 
+encodeBase64Url(Bs) when is_list(Bs) ->
+    encodeBase64Url(list_to_binary(Bs), <<>>);
+encodeBase64Url(Bs) ->
+    encodeBase64Url(Bs, <<>>).
+    
+encodeBase64Url(<<B:3/binary, Bs/binary>>, Acc) ->
+    <<C1:6, C2:6, C3:6, C4:6>> = B,
+    encodeBase64Url(Bs, <<Acc/binary, (encUrl(C1)), (encUrl(C2)), (encUrl(C3)), (encUrl(C4))>>);
+encodeBase64Url(<<B:2/binary>>, Acc) ->
+    <<C1:6, C2:6, C3:6, _:6>> = <<B/binary, 0>>,
+    <<Acc/binary, (encUrl(C1)), (encUrl(C2)), (encUrl(C3))>>;
+encodeBase64Url(<<B:1/binary>>, Acc) ->
+    <<C1:6, C2:6, _:12>> = <<B/binary, 0, 0>>,
+    <<Acc/binary, (encUrl(C1)), (encUrl(C2))>>;
+encodeBase64Url(<<>>, Acc) ->
+    Acc.
+
 %%
 %% decodeBase64(BinaryChars) -> Binary
 %%
@@ -265,6 +297,23 @@ decode1(<<C1, C2, C3, C4, Cs/binary>>, Acc) ->
 decode1(<<>>, Acc) ->
     Acc.
 
+decodeBase64Url(Cs) when is_list(Cs)->
+    decodeBase64Url(list_to_binary(Cs));
+decodeBase64Url(Cs) ->
+    decode1Url(Cs, <<>>).
+
+decode1Url(<<C1, C2>>, Acc) ->
+    <<B1, _:16>> = <<(decUrl(C1)):6, (decUrl(C2)):6, 0:12>>,
+    <<Acc/binary, B1>>;
+decode1Url(<<C1, C2, C3>>, Acc) ->
+    <<B1, B2, _:8>> = <<(decUrl(C1)):6, (decUrl(C2)):6, (decUrl(C3)):6, (decUrl(0)):6>>,
+    <<Acc/binary, B1, B2>>;
+decode1Url(<<C1, C2, C3, C4, Cs/binary>>, Acc) ->
+    Bin = <<Acc/binary, (decUrl(C1)):6, (decUrl(C2)):6, (decUrl(C3)):6, (decUrl(C4)):6>>,
+    decode1Url(Cs, Bin);
+decode1Url(<<>>, Acc) ->
+    Acc.
+
 %% enc/1 and dec/1
 %%
 %% Mapping: 0-25 -> A-Z, 26-51 -> a-z, 52-61 -> 0-9, 62 -> +, 63 -> /
@@ -275,6 +324,15 @@ enc(C) ->
 dec(C) ->
     62*?st(C,43) + ?st(C,47) + (C-59)*?st(C,48) - 69*?st(C,65) - 6*?st(C,97).
 
+%% encUrl/1 and decUrl/1
+%%
+%% Mapping: 0-25 -> A-Z, 26-51 -> a-z, 52-61 -> 0-9, 62 -> -, 63 -> _
+%%
+encUrl(C) ->
+    65 + C + 6*?st(C,26) - 75*?st(C,52) -13*?st(C,62) + 49*?st(C,63).
+
+decUrl(C) ->
+    62*?st(C,45) + (C-58)*?st(C,48) - 69*?st(C,65) + 33*?st(C,95) - 39*?st(C,97).
 
 dict_find(Key, Dict, DefaultValue) ->
     case dict:find(Key, Dict) of
@@ -283,3 +341,57 @@ dict_find(Key, Dict, DefaultValue) ->
     error ->
         DefaultValue
     end.
+
+
+file_read_size(FileName) ->
+    case file:read_file_info(FileName) of
+        {ok, FileInfo} ->
+            FileInfo#file_info.size;
+        Error -> Error
+    end.
+
+to_binary(V) when is_binary(V) ->
+    V;
+to_binary(V) when is_list(V) ->
+    try
+        list_to_binary(V)
+    catch
+        _ ->
+            list_to_binary(io_lib:format("~p", [V]))
+    end;
+to_binary(V) when is_atom(V) ->
+    list_to_binary(atom_to_list(V));
+to_binary(V) ->
+    list_to_binary(io_lib:format("~p", [V])).
+
+to_list(V) when is_list(V) ->
+    V;
+to_list(V) when is_binary(V) ->
+    binary_to_list(V);
+to_list(V) when is_atom(V) ->
+    atom_to_list(V);
+to_list(V) ->
+    lists:flatten(io_lib:format("~p", [V])).
+
+url_encode(Bin) when is_binary(Bin) ->
+    url_encode(binary_to_list(Bin));
+url_encode([H|T]) ->
+    if
+    H >= $a, $z >= H ->
+        [H|url_encode(T)];
+    H >= $A, $Z >= H ->
+        [H|url_encode(T)];
+    H >= $0, $9 >= H ->
+        [H|url_encode(T)];
+    H == $_; H == $.; H == $-; H == $: ->
+        [H|url_encode(T)];
+    true ->
+        case lists:flatten(io_lib:format("~.16.0B", [H])) of
+        [X, Y] ->
+            [$%, X, Y | url_encode(T)];
+        [X] ->
+            [$%, $0, X | url_encode(T)]
+        end
+    end;
+url_encode([]) ->
+    [].
\ No newline at end of file
