@@ -12,7 +12,6 @@
 -export([shell_quote/1, cmd/1, cmd_string/1, cmd_port/2]).
 -export([record_to_proplist/2, record_to_proplist/3]).
 -export([safe_relative_path/1, partition/2]).
--export([to_lower/1]).
 -export([test/0]).
 
 -define(PERCENT, 37).  % $\%
@@ -239,7 +238,7 @@ urlsplit_scheme(Url) ->
 urlsplit_scheme([], Acc) ->
     {"", lists:reverse(Acc)};
 urlsplit_scheme(":" ++ Rest, Acc) ->
-    {to_lower(lists:reverse(Acc)), Rest};
+    {string:to_lower(lists:reverse(Acc)), Rest};
 urlsplit_scheme([C | Rest], Acc) ->
     urlsplit_scheme(Rest, [C | Acc]).
 
@@ -391,11 +390,11 @@ parse_header(String) ->
                         %% Skip anything with no value
                         Acc;
                     {Name, [$\= | Value]} ->
-                        [{to_lower(string:strip(Name)),
+                        [{string:to_lower(string:strip(Name)),
                           unquote_header(string:strip(Value))} | Acc]
                 end
         end,
-    {to_lower(Type),
+    {string:to_lower(Type),
      lists:foldr(F, [], Parts)}.
 
 unquote_header("\"" ++ Rest) ->
@@ -418,7 +417,7 @@ record_to_proplist(Record, Fields) ->
     record_to_proplist(Record, Fields, '__record').
 
 %% @spec record_to_proplist(Record, Fields, TypeKey) -> proplist()
-%% @doc Return a proplist of the given Record with each field in the 
+%% @doc Return a proplist of the given Record with each field in the
 %%      Fields list set as a key with the corresponding value in the Record.
 %%      TypeKey is the key that is used to store the record type
 %%      Fields should be obtained by calling record_info(fields, record_type)
@@ -438,20 +437,6 @@ shell_quote([C | Rest], Acc) when C =:= $\" orelse C =:= $\` orelse
 shell_quote([C | Rest], Acc) ->
     shell_quote(Rest, [C | Acc]).
 
-to_lower_char(C) when is_integer(C),  C >= $A, C =< $Z ->
-    C + 32;
-to_lower_char(C) when is_integer(C),  C >= 16#C1, C =< 16#D6 ->
-    C + 32;
-to_lower_char(C) when is_integer(C),  C >= 16#D8, C =< 16#DE ->
-    C + 32;
-to_lower_char(C) ->
-    C.
-
-to_lower(S) when is_list(S) ->
-    [to_lower_char(C) || C <- S];
-to_lower(C) when is_integer(C) ->
-    to_lower_char(C).
-
 test() ->
     test_join(),
     test_quote_plus(),
